
open Vcgen
open Cmdliner

open Terms

let process_output lines =
  List.for_all (fun s -> s = "valid") lines
  
    

(* Tests of the form  (pctxt,t,rk) *)
let test1 =
  let ctxt =
    [("r1",KRefine("r",KRec,Not(REmpty(Id("r")))));
     ("r2",KRefine("r",KRec,And(Not(In(RLabel(Id("r1")),LabSet(Id("r2"))))
          ,And(Not(REmpty(Id("r"))),LabApart(Id("r1"),Id("r"))))))]
  in 
  let typ = TRecCons(RLabel(Id("r1")),RHead(Id("r1")),
                 TRecCons(RLabel(Id("r2")),RHead(Id("r2")),TEmptyRec))
  in
  let rk = trivial_refine KRec in
  (ctxt,typ,rk)

let checker verb no_check ext =
  if (not ext) then
    Vcgen.extensionality_mode := true
  else
    Vcgen.extensionality_mode := false ;
  let test_kinding = [test1] in
  let test_subkinding = [] in
  let vcs = gen_vcs test_kinding test_subkinding in
  let out = ref [] in
  if (not no_check) then
    out := run_cvc4 vcs ;
     if (verb) then
      begin
        Printf.printf "Verbose mode, printing CVC VCs file:\n%s\n" vcs ;
        if (!out <> []) then
        begin
          print_endline "CVC4 Output:" ;
          Printf.printf "\n%s\n" (String.concat "\n" !out)
        end
         
      end
     else
    Printf.printf "Non-verbose Output: %B \n" (process_output !out) ;
    Printf.printf "All done.\n"

(* Cmdline stuff *)
let verbose =
  let doc = "Verbose output." in
  Arg.(value & flag & info ["v";"verbose"] ~doc)

let extensionality =
  let doc = "Extensionality principle for higher-order kinds." in
  Arg.(value & flag & info ["e";"extkind"] ~doc)

let no_check =
  let doc = "Disable checking of VCs." in
  Arg.(value & flag & info ["n";"no_check"] ~doc)

let cmd =
  let doc = "Generate VCs and check them
  using the cvc4 binary." in
  let man = [
    `S Manpage.s_description;
    `P "Stub for now.";
    `P "More stub.";
    `S Manpage.s_bugs; `P "Report bugs to me."  ]
  in
  Term.(const checker $ verbose $ no_check $ extensionality),
  Term.info "checker" ~version:"v0.1b" ~doc ~exits:Term.default_exits ~man

(****)
    
    
let () = Term.(exit @@ eval cmd)
