{
 open Parser;;
 let count = ref 0;;
}

rule token = parse
  [' ' '\t' '\n' '\r']  { token lexbuf }
| "/*"             { incr count; comment lexbuf }
| "//"             { linecomment lexbuf } 

| eof   		  { EOF }

| "quit"          { QUIT } 
| "eval"          { EVAL } 
| "type"          { CTYPE } 
| "kind"          { CKIND } 
| "ok"            { OKC } 
| "bool"          { BOOLT }
| "int"           { INTT }
| "unit"          { UNITT }
| "string"        { STRINGT }
| "top"           { TOPT }
| "bot"           { BOTT }
| "ref"           { REFT }
| "new"           { NEW }
| "let"           { LET }
| "letrec"        { LETREC }
| "val"           { VAL }
| "end"           { END }

| "Type"          { KTYPE }
| "Fun"           { KFUN }
| "Rec"           { KREC }
| "Lab"           { KLAB }
| "Ref"           { KREF }
| "Gen"           { KGEN }

| "fun"           { LFUN }
| "true"          { TRUE }
| "false"         { FALSE }
| "if"            { IF }
| "then"          { THEN }
| "else"          { ELSE }
| "true"          { INTT }
| "head"          { HEAD }     
| "tail"          { TAIL }  
| "headlb"        { HLAB }  
| "empty"	  { EMP  }
| "labSet"	  { LABS  }
| "in"	          { IN  }
| "inl"	          { INL }

| "dom"           { DOM }     
| "cod"           { COD }   

| "refof"         { REFOF }   

| "trace"         { TRACE }   

| "Pi"            { PI }     
| "All"           { FORALL }   
| "+" 		  { PLUS }
| "-" 		  { MINUS }
| "*" 		  { TIMES }
| "/" 		  { DIV }                    
| "," 		  { COMMA }
| "." 		  { DOT }
| "::" 		  { DCOLON }
| ":" 		  { COLON }
| ";;" 	 	  { TERM }
| "(" 		  { LPAR }
| ")" 		  { RPAR }
| "[" 		  { LBRA }
| "]" 		  { RBRA }
| "[|" 		  { TLBRA }
| "|]" 		  { TRBRA }
| "<" 		  { LT }
| ">" 		  { GT }
| "{" 		  { LCPAR }
| "}" 		  { RCPAR }
| "=" 		  { EQ }
| "==" 		  { EQQ }
| "&&" 		  { AND }
| "U" 		  { UNION }
| "||" 		  { OR }
| "~" 		  { NOT }
| "=>" 		  { IMPLIES }
| ":=" 		  { ASS }
| "=>" 		  { IMPLIES }
| "@" 		  { ATSGN }
| "`" 		  { HASH }
| "->" 	          { RARROW }
| "|" 	  	  { VBAR }
| "!" 	  	  { DEREF }
| "^" 	  	  { STRCAT }
| "#" 	  	  { LABA }
| ['0'-'9']+                      { INT(int_of_string (Lexing.lexeme lexbuf)) }
| ['0'-'9']* "." ['0'-'9']*       { FLOAT(float_of_string (Lexing.lexeme lexbuf)) }
| "\"" ['A'-'Z' 'a'-'z' '0'-'9' '_' '<' '>' '\\' ' ' '*' '-']* "\""  { let s = Lexing.lexeme lexbuf in STRING(s) }
| ['A'-'Z' 'a'-'z' '0'-'9' '_']*  { ID(Lexing.lexeme lexbuf) }

| _ { raise Parsing.Parse_error }

and comment = parse
  ['\t' '\n']                 { comment lexbuf }
| "*/"                        { decr count; if !count = 0 then token lexbuf else comment lexbuf }
| "/*"                        { incr count; comment lexbuf }
| [' '-')' '+'-'.' '0'-'~']*  { comment lexbuf }
| '*'+ [' '-'.' '0'-'~']      { comment lexbuf }
| '/'+ [' '-')' '+'-'~']      { comment lexbuf }

and linecomment = parse
  ['\n']                      { token lexbuf }
| _                           { linecomment lexbuf }
 
