open Syntax
open Terms
open Comm
open Cmdliner
open Lexing

let wrapper = ref (Some ["rlwrap";"ledit"])
(**let wrapper = ref None**)

let rec top_level lexbuf =
	(
   try
        print_string "# " ;
        flush stdout;
        let s = Parser.main Lexer.token lexbuf
        in (Terms.cc := 0;
  	      match s with
  	        Quit -> exit 0
 		|  Eval(a) -> 
		     print_term a;
                     print_string "\n";
                     print_string "value: ";
                     print_term (evaluate_term a []);
                     print_string "\n"
		|  Typec(a,et) -> (
                          let tt = typecheck [] [] a et KType false [] in
                          let _ = typecheck [] [] tt KType KindOK false [] in
                          (  
                             print_string " - : "; 
                             print_term tt; 
                             print_string " = "; print_term (evaluate_term a []); print_string "\n"))
		|  Kindc(a,ek) ->
                          let kk = typecheck [] [] a ek KindOK false [] in
                          ( print_string " - :: "; 
                            print_term kk; 
                            print_string " = ";print_term (evaluate_term a []); print_string "\n")
		|  Okc(a) ->
                          let kk = typecheck [] [] a KindOK KindOK false [] in
                          (print_term kk; print_string "\n"))

   with
   
        Parsing.Parse_error -> print_string "Syntax error\n" 
      
      | Failure(s) -> print_endline s
      | Sys.Break -> Vcgen.clean_up_cvc_proc () ; exit 0                       
    );
      top_level lexbuf 
;;

let main no_ext no_wrap debug file =
    Vcgen.debug := debug ; 
  (if no_wrap then
    wrapper := None);
  (if no_ext then
    Vcgen.extensionality_mode := false) ;
  begin match !wrapper with
    | None -> ()
    | Some lst ->
      Sys.catch_break false ;
      let n = Array.length Sys.argv + 2 in
      let args = Array.make n "" in
      Array.blit Sys.argv 0 args 1 (n - 2) ;
        args.(n - 1) <- "--no-wrapper" ;
        List.iter
          (fun wrapper ->
             try
               args.(0) <- wrapper ;
               Unix.execvp wrapper args
             with Unix.Unix_error _ -> ())
          lst
  end ;
    Sys.catch_break true ; Vcgen.init_cvc_proc () ; at_exit Vcgen.clean_up_cvc_proc ;
    let inchan = if (Sys.file_exists file) then (open_in file) else stdin in
    let lexbuf = Lexing.from_channel (inchan) in top_level lexbuf
    




(* Cmdline stuff *)
let file =
  let doc = "Load a file." in
  Arg.(value & pos 0 string "" & info [] ~docv:"FILE" ~doc)

let no_wrap =
  let doc = "Command line wrapper utility." in
  Arg.(value & flag & info ["nw"; "no-wrapper"] ~doc)

let debug =
  let doc = "Debug mode." in
  Arg.(value & flag & info ["d"; "debug"] ~doc)

let no_ext =
  let doc = "Disable the extensionality principle for higher-order kinds." in
  Arg.(value & flag & info ["e";"extkind"] ~doc)

let cmd =
  let doc = "Refinement Kind Checker." in
  let man = [
    `S Manpage.s_description;
    `P "Stub for now.";
    `P "More stub.";
    `S Manpage.s_bugs; `P "Report bugs to me."  ]
  in
  Term.(const main $ no_ext $ no_wrap $ debug $ file),
  Term.info "checker" ~version:"v0.1b" ~doc ~exits:Term.default_exits ~man

(****)
    
    
let () = Term.(exit @@ eval cmd)
