%{
open Syntax;;
open Terms;;
open Comm;;
%}

%token EOF 
%token QUIT
%token EVAL
%token CTYPE
%token CKIND
%token OKC

%token LPAR
%token RPAR

%token LBRA
%token RBRA
%token LCPAR
%token RCPAR
%token TLBRA
%token TRBRA
%token LT
%token GT
%token ATSGN
%token EQQ
%token EQ
%token AND
%token OR
%token NOT
%token IMPLIES
%token LABS
%token UNION
%token LABA
%token ASS
%token DEREF
%token IN
%token INL
%token STRCAT

%token LET
%token LETREC
%token VAL
%token END


%token COMMA
%token COLON
%token DCOLON
%token DOT
%token VBAR

%token TERM

%token RARROW
%token LRARROW

%token HEAD
%token TAIL
%token HLAB
%token EMP

%token NEW

%token DOM
%token COD

%token BOOLT
%token INTT
%token UNITT
%token TOPT
%token BOTT
%token STRINGT

%token REFT
%token REFOF

%token KTYPE
%token KFUN
%token KREC
%token KLAB
%token KREF
%token KGEN

%token LFUN
%token HFUN

%token PI
%token FORALL

%token IF
%token THEN
%token ELSE
%token TRUE
%token FALSE

%token PLUS 
%token MINUS
%token TIMES
%token DIV
%token NEG

%token HASH
%token ID


%token TRACE


%token <int> INT
%token <float> FLOAT
%token <string> STRING
%token <string> ID

%type <Comm.comm> main
%type <Comm.comm> sline
%type <Syntax.term> expr
%type <Syntax.term> term
%type <Syntax.term> factor
%type <Syntax.term> recordvl
%type <Syntax.term> recordtl
%type <Syntax.term> vrecordc
%type <Syntax.term> trecordc

%left PLUS MINUS
%left TIMES DIV
%left NEG
%left EMP

%start main

%%

main:
sline TERM  { $1 }

sline:
| QUIT { Quit }
|  CTYPE expr { Typec($2,Ttop) }
|  CTYPE expr COLON expr { Typec($2,$4) }
|  CKIND expr { Kindc($2,KType) }
|  CKIND expr DCOLON expr { Kindc($2,$4) }
|  OKC expr { Okc($2) }
|  expr { Eval ($1) }

expr:
| term { $1 }
| term RARROW term { TArrow($1,$3) }
| expr OR term { Or($1,$3) }
| expr PLUS term { NumPlus($1,$3) } 
| expr IMPLIES term { Implies($1,$3) }
| expr ASS term { Assign($1,$3) }
| expr term { App($1,$2) }


term:
| factor { $1 }
| factor EQQ factor { Equals($1,$3,KType) } 
| term TIMES factor { NumTimes($1,$3) } 
| term UNION factor { LabUnion($1,$3) }
| term AND factor { And($1,$3) }
| term STRCAT factor { StrConcat($1,$3) }
| term LABA factor { LabApart($1,$3) }
| term INL factor  { In($1,$3) }  


factor:

| LFUN ID COLON term RARROW factor { Lam ($2,$4,$6) }
| LFUN ID DCOLON term RARROW factor { TLam ($2,$4,$6) }
| IF expr THEN expr ELSE factor { Ite ($2,$4,$6) }
| FORALL ID DCOLON term DOT factor { TForall ($2,$4,$6) }
| PI ID DCOLON term DOT factor { KPi ($2,$4,$6) }
| LET ID COLON expr EQ expr IN expr END { Letval ($2,$4,$6,$8) }
| LET ID DCOLON expr EQ expr IN expr END { Lettype ($2,$4,$6,$8) }
| LETREC ID COLON expr EQ expr IN expr END { Letrecval ($2,$4,$6,$8) }
| LETREC ID DCOLON expr EQ expr IN expr END { Letrectype ($2,$4,$6,$8) }


| HASH ID { Label($2) }
| ID { Id($1) }
| INT { Num($1) }
| STRING { let s = $1 in String(String.sub s 1 ((String.length s)-2))  }

| BOOLT { TBool }
| INTT { TInt }
| UNITT { TUnit }
| TOPT { Ttop }
| BOTT { Tbot }
| STRINGT { TString }

| REFT factor { TRef ($2) }
| REFOF factor { TRefOf ($2) }

| KGEN factor { KGen ($2) }


| KTYPE { KType }
| KREC { KRec }
| KREF { KRef }
| KFUN { KFun }
| KLAB { KLab }

| TRUE { Bool true }
| FALSE { Bool false }

| HEAD factor  { RHead($2) }
| TAIL factor  { RTail($2) }
| HLAB factor  { RLabel($2) }
| DOM factor   { TDom($2) }
| COD factor   { TCodom($2) }

| NEW factor   { Ref($2) }
| DEREF factor   { Deref($2) }

| NOT factor  { Not ($2) }
| LABS factor { LabSet $2 }

| EMP factor   { REmpty($2) }

| LCPAR ID DCOLON expr VBAR expr RCPAR { KRefine ($2,$4,$6) }
| recordvl  { (revVR $1) }
| recordvl ATSGN factor  { recVAT (revVR $1) ( $3) }
| recordtl  {  (revTR $1) }
| recordtl ATSGN factor  { recTAT (revTR $1) ( $3) }

| LPAR expr RPAR { $2 }

recordvl: 
| LBRA vrecordc RBRA {  $2 }

recordtl: 
| TLBRA  trecordc  TRBRA {  $2 }

vrecordc:
| { EmptyRec }
| expr EQ expr { RecCons($1,$3,EmptyRec) }
| vrecordc COMMA expr EQ expr { RecCons($3,$5,$1) }

trecordc:
| { TEmptyRec }
| expr COLON expr { TRecCons($1,$3,TEmptyRec) }
| trecordc COMMA expr COLON expr { TRecCons($3,$5,$1) }
