
type term = 

       | Id of string
	| Num of int
	| Bool of bool
        | String of string
	| Lam of string * term * term  
	| App of term * term 
        | Letval of string * term * term * term
        | Lettype of string * term * term * term
        | Letrecval of string * term * term * term
        | Letrectype of string * term * term * term
	| Clos of string * term * termenv
	| FClos of string * term * termenv
	| Ite of term * term * term
     | Ref of term
     | Deref of term
     | Assign of term * term
     | Loc of (term ref)
     | Kindcase of term * term * string * term * term
     | EmptyRec
	| RecCons of term * term * term
	| RLabel of term
	| RHead of term
	| RTail of term
	| REmpty of term

	| TInt
	| TBool
	| TString
	| TUnit
	| Ttop
	| Tbot

	| TLam of string * term * term
 	| TClos of string * term * termenv
	| TForall of string * term * term
	| TMap of term * term

     | Label of string

    	| TEmptyRec
	| TRecCons of term * term * term
	| TArrow of term * term
	| TDom of term
	| TCodom of term
	| TRef of term
	| TRefOf of term

	| TIte of term * term * term 

     | KRec
     | KFun
     | KType
     | KRef
     | KLab
	| KGen of term

	| KPi of string * term * term
     | KRefine of string * term * term

     | EqVal of term * term
     | NumTimes of term * term
     | NumPlus of term * term
     | NumMinus of term * term

     | StrConcat of term * term


     | Implies of term * term
     | And of term * term
     | Not of term 
     | Equiv of term * term * term
     | Equals of term * term * term
     | Or of term * term
     | LabSet of term
     | LabApart of term * term
     | In of term * term
     | LSet
     | LabConcat of term * term
     | LabUnion of term * term
     | LabSetConcat of term * term                

       | KindOK

  | KPiList of (string * term) list * term
  | AppList of term * term list

  | RecF of string * term * term

  | Tracer of term

and termenv = (string * term) list

and ihenv = (string * term * term) list

let rec print_term0 t lev =
  match t with

  |  Bool true -> print_string "true"
  |  Bool false -> print_string "false"
  |  Num n -> print_int n
  |  String s -> print_string ("\""^s^"\"")

  | KType -> print_string "Type"
  | KFun -> print_string "Fun"
  | KRec -> print_string "Rec"
  | KRef -> print_string "Ref"
  | KLab -> print_string "Lab"

  | KGen k -> ( print_string "Gen ";
		print_term0 k 0
              )

  | KRefine (x,k,p) -> (
		print_string "{ ";
		print_string x;
		print_string " :: ";
		print_term0 k 0;
		print_string " | ";
		print_term0 p 0;
		print_string " }")

  | Letval (x,t,d,b) -> (
		print_string "let ";
		print_string x;
		print_string " : ";
		print_term0 t 0;
		print_string " = ";
		print_term0 d 0;
		print_string " in ";
		print_term0 b 0 )

  | Lettype (x,t,d,b) -> (
		print_string "let ";
		print_string x;
		print_string " :: ";
		print_term0 t 0;
		print_string " = ";
		print_term0 d 0;
		print_string " in ";
		print_term0 b 0 )


  | Letrecval (x,t,d,b) -> (
		print_string "letrec ";
		print_string x;
		print_string " : ";
		print_term0 t 0;
		print_string " = ";
		print_term0 d 0;
		print_string " in ";
		print_term0 b 0 )

  | Letrectype (x,t,d,b) -> (
		print_string "letrec ";
		print_string x;
		print_string " :: ";
		print_term0 t 0;
		print_string " = ";
		print_term0 d 0;
		print_string " in ";
		print_term0 b 0 )

  | KPi (x,k,p) -> (
		print_string "Pi ";
		print_string x;
		print_string " :: ";
		print_term0 k 0;
		print_string ".";
		print_term0 p 0)

  | TInt -> print_string "int"
  | TString -> print_string "string"
  | TBool -> print_string "bool"
  | TUnit -> print_string "unit"
  | Ttop -> print_string "top"
  | Tbot -> print_string "bot"

  | KindOK -> print_string "Kind"


  |  Ite(c,t,e) -> (
		print_string "if ";
		print_term0 c 0;
		print_string " then ";
		print_term0 t 0;
		print_string " else ";
		print_term0 e 1)
  |  Id s -> (print_string (s);
		print_string "")
  |  Label l -> (print_string "`"; print_string l)
  |  RecF (f,t,b) -> (
		print_string "(rec ";
		print_string f;
		print_string ":";
		print_term0 t 0;
		print_string " == ";
		print_term0 b 0;
		print_string ")"
	      )
  |  Lam (x,u,b) -> (
		if lev==1 then print_string "(";
		print_string "fun ";
		print_string x;
		print_string ":";
		print_term0 u 0;
		print_string " -> ";
		print_term0 b 1;
		if lev==1 then print_string ")"
	      )
  |  TLam (x,k,b) -> (
		if lev==1 then print_string "(";
		print_string "fun ";
		print_string x;
		print_string "::";
		print_term0 k 0;
		print_string " -> ";
		print_term0 b 1;
		if lev==1 then print_string ")";
		)
  |  TForall (x,k,b) -> (
		print_string "All ";
		print_string x;
		print_string "::";
		print_term0 k 0;
		print_string " . ";
		print_term0 b 1)
  |  Clos (x,b,e) -> (
		print_string "< ";
		print_string x;
		print_string "| ";
		print_term0 b 0;
		print_string "> ")
  |  TClos (x,b,e) -> (
		print_string "< ";
		print_string x;
		print_string "| ";
		print_term0 b 0;
		print_string "> ")
  |  FClos (x,b,e) -> (
		print_string "> ";
		print_string x;
		print_string "| ";
		print_term0 b 0;
		print_string "> ")
  |  App (u,v) -> (
		if lev==1 then print_string "(";
		print_term0 u 1;
		print_string " ";
		print_term0 v 0;
		if lev==1 then print_string ")"
	)
  |  TMap (u,v) -> (
		if lev==1 then print_string "tmap(";
		print_term0 u 1;
		print_string " ";
		print_term0 v 0;
		if lev==1 then print_string ")"
	)
  |  TArrow (u,v) -> 
		print_term0 u 0;
		print_string " -> ";
		print_term0 v 0
  |  TDom e ->
		print_string "dom(";
		print_term0 e 0;
		print_string ")";
  |  TCodom e ->
		print_string "codom(";
		print_term0 e 0;
		print_string ")";

  |  TRef e -> 
	      print_string "ref ";
		print_term0 e 0;
  |  TRefOf e ->
		print_string "refOf ";
		print_term0 e 0;

  | EmptyRec -> print_string "[]"
  | TEmptyRec -> print_string "[||]"

  | RecCons(l,v,r)  -> print_string "["; print_recv l v r
  | TRecCons(l,v,r) -> print_string "[|"; print_rect l v r

  | RLabel e ->
		print_string "headlb(";
		print_term0 e 0;
		print_string ")"
  | RHead e -> 
		print_string "head(";
		print_term0 e 0;
		print_string ")"
  | RTail e ->
		print_string "tail";
		print_string "(";
		print_term0 e 0;
		print_string ")"
  | REmpty e ->
		print_string "empty(";
		print_term0 e 0;
		print_string ")"

  | Ref a -> 
		print_string "new(";
		print_term0 a 0;
		print_string ")"

  | Deref a -> print_string "!";
		print_term0 a 1

  | Assign (e1,e2) -> 
		print_term0 e1 1;
		print_string " := ";
		print_term0 e2 1;
  | Loc a -> 
		print_string "loc(";
		print_term0 !a 0;
		print_string ")"
  | Equals (a, b, c) ->
		print_string "(";
		print_term0 a 0;
		print_string " == "; 
		print_term0 b 0;
		print_string ":";
		print_term0 c 0;
		print_string ")"
  | And (a, b) ->
		print_string "(";
		print_term0 a 0;
		print_string " && "; 
		print_term0 b 0;
		print_string ")"
  | Or (a, b) ->
		print_string "(";
		print_term0 a 0;
		print_string " || "; 
		print_term0 b 0;
		print_string ")"
  | Implies (a, b) ->
		print_string "(";
		print_term0 a 0;
		print_string " => "; 
		print_term0 b 0;
		print_string ")"
  | Not a ->
		print_string "~";
		print_term0 a 0
  | LabSet a ->
		print_string "labSet ";
		print_string "(";
		print_term0 a 0;
		print_string ")"
  | In (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " inl "; 
		print_term0 b 0;
		print_string ")"
  | LabApart (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " # "; 
		print_term0 b 0;
		print_string ")"
  | LabUnion (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " U "; 
		print_term0 b 0;
		print_string ")"
  | NumPlus (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " + "; 
		print_term0 b 0;
		print_string ")"
  | NumMinus (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " - "; 
		print_term0 b 0;
		print_string ")"
  | NumTimes (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " * "; 
		print_term0 b 0;
		print_string ")"

  | StrConcat (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string "^"; 
		print_term0 b 0;
		print_string ")"

  | EqVal (a,b) ->
		print_string "(";
		print_term0 a 0;
		print_string " = "; 
		print_term0 b 0;
		print_string ")"

  | LSet -> print_string "LSet"

  | Tracer (a) ->
		print_term0 a lev;

  | _  -> failwith ("Cannot print term!");


and  print_recv l v r  =
	print_term0 l 0; 
	print_string "=";
	print_term0 v 0;
	match r with
	EmptyRec -> print_string "]"
	| RecCons(l',v',r') -> 
			print_string ", ";
			print_recv l' v' r'
	| _ -> (print_string "]@"; print_term0 r 0)
and  print_rect l v r  =
	print_term0 l 0;
	print_string ":";
	print_term0 v 0;
	match r with
	TEmptyRec -> print_string "|]"
	| TRecCons(l',v',r') -> 
			print_string ", ";
			print_rect l' v' r'
	| _ -> (print_string "|]@"; print_term0 r 0)
;;

let print_term t = print_term0 t 0
;;