open String
open Syntax
open Vcgen

let cc = ref 0;;

let new_id =
   fun () -> (cc := !cc + 1;"X$"^(string_of_int !cc));;

let pause_ref = ref false;;

let set_pause_mode () = pause_ref := true;;

let pause () = if !pause_ref then (input_char stdin; ()) else ();;

let count_down = ref 1;;

let count_debug () =
if (!count_down) > 0 then (count_down:=(!count_down)-1;false) else (Vcgen.debug := true; true);;

let rec print_env e =
  match e with
   |    [] -> print_string "\n"
   |   (x,t)::e' -> (print_string (x^" : "); print_term t ; print_string " ,\n"; print_env e')
;;

let rec print_ih e =
  match e with
   |    [] -> print_string "\n"
   |   (x,t1,t2)::e' -> (print_string ("( "^x^" , "); print_term t1 ; print_string " , "; print_term t2 ; print_string ") , "; print_ih e')
;;


let andPrim a b =
  match a,b with
  | Bool true, Bool true -> Bool true
  |  _,_ -> Bool false;;

let orPrim a b =
  match a,b with 
  | Bool false, Bool false -> Bool false
  | _,_ -> Bool true;;

let impliesPrim a b =
  match a,b with 
  | Bool true, Bool false -> Bool false
  | Bool _, Bool _ -> Bool true;;

let notPrim a =
  match a with 
  | Bool false -> Bool true
  | Bool true -> Bool false;;

let timesNumPrim a b =
  match a,b with 
  | Num x, Num y -> Num (x*y);;

let plusNumPrim a b =
  match a,b with 
  | Num x, Num y -> Num (x+y);;

let minusNumPrim a b =
  match a,b with 
  | Num x, Num y -> Num (x-y);;

let strConcatPrim a b =
  match a,b with 
  | String x, String y -> String (x^y);;

let eqPrim a b =
  match a,b with 
  | Bool x, Bool y -> Bool(x = y)
  | Num x, Num y -> Bool(x = y);;

let revVR t = 
  let rec aux t' a = 
	match t' with 
 	| RecCons (l,v,r) -> aux r (RecCons (l,v,a))
        | EmptyRec -> a
   in aux t EmptyRec
;;

let revTR t = 
  let rec aux t' a = 
	match t' with 
	| TRecCons (l,v,r) -> aux r (TRecCons (l,v,a))
        | TEmptyRec -> a
   in aux t TEmptyRec
;;


let recTAT t q = 
  let rec aux t' a = 
	match t' with 
	| TRecCons (l,v,r) -> TRecCons (l,v,aux r q)
        | _ -> a
   in aux t q
;;

let recVAT t q = 
  let rec aux t' a = 
	match t' with 
	| RecCons (l,v,r) -> RecCons (l,v,aux r q)
        | _ -> a
   in aux t q
;;

let rec find id env =
	match env with
           [] -> failwith ("env_err:"^id^" not found")
   	| (i,t)::r -> if i=id then t else find id r;;

let rec evaluate_term t env =
  match t with

  | Bool _ -> t
  | Num _ -> t
  | String _ -> t

  | TInt -> t
  | TBool -> t
  | TString -> t
  | TUnit -> t
  | Ttop -> t
  | Tbot -> t


  | KType -> t
  | KFun -> t
  | KRec -> t
  | KRef -> t
  | KLab -> t
  | KGen k -> KGen (evaluate_term k env)
  | KRefine (x,k,p) -> KRefine(x,(evaluate_term k env),(evaluate_term k ((x,k)::env)))

  |  KPi (x,u,b) -> t (* see later eval to closure ? *)

  |  Ite(c,t,e) ->
	 let v1 = evaluate_term c env in
       (match v1 with
          Bool true -> evaluate_term t env
	 |  Bool false -> evaluate_term e env
	 |  _ -> failwith "bool_err")
  |  Id s -> find s env
  |  Lam (x,_,b) -> Clos (x,b,env)
  |  Clos _ -> t
  |  TLam (x,_,b) -> TClos (x,b,env)
  |  TClos _ -> t
  |  TForall (x,_,b) -> FClos (x,b,env) 
  |  FClos _ -> t
  |  Letval (x,d,t,b) -> 
       let v1 = (evaluate_term t env) in
            evaluate_term b ((x,v1)::env) 
  |  Lettype (x,d,t,b) -> 
       let v1 = (evaluate_term t env) in
            evaluate_term b ((x,v1)::env) 
  |  Letrecval (f,d,t,b) -> 
       let v1 = evaluate_term t ((f,Id f)::env) in evaluate_term b ((f,RecF(f,d,v1))::env)
  |  Letrectype (f,d,t,b) -> 
       let v1 = evaluate_term t ((f,Id f)::env) in evaluate_term b ((f,RecF(f,d,v1))::env)
  |  App (u,v) -> 
       let v1 = (evaluate_term u env) and
           v2 = (evaluate_term v env) in 
       ( match v1 with 
          Lam(x,_,b) -> evaluate_term b ((x,v2)::env)
        | TLam(x,_,b) -> evaluate_term b ((x,v2)::env)
        | Clos (x,b,e) -> evaluate_term b ((x,v2)::e)
        | TClos (x,b,e) -> evaluate_term b ((x,v2)::e)
        | RecF (f, t, z) -> (match z with 
                              | Lam  (x,_,b) -> evaluate_term b ((x,v2)::env) 
                              | Clos (x,b,e) -> evaluate_term b ((x,v2)::(f,RecF (f,t,z))::e)
                              | TLam  (x,_,b) -> evaluate_term b ((x,v2)::env) 
                              | TClos (x,b,e) -> evaluate_term b ((x,v2)::(f,RecF (f,t,z))::e)
                               | _ -> failwith "No recursive value in app") 
        | _ -> failwith "app_err"
       )
  |  TMap (u,v) -> 
           let v1 = (evaluate_term u env) and
               v2 = (evaluate_term v env) in 
           (match v1 with 
              | TForall (x,_,b) -> evaluate_term b ((x,v2)::env)
              | FClos (x,b,e) -> evaluate_term b ((x,v2)::e)
              | _ -> failwith "tmap_err"
       )
  |  TArrow (u,v) -> 
       let d = (evaluate_term u env) and
           c = (evaluate_term v env) in 
       	TArrow(d,c)
  | Label s -> t
  | EmptyRec -> t
  | TEmptyRec -> t
  | RecCons(l,v,r) -> 
       let d = (evaluate_term r env) and
           c = (evaluate_term v env) and 
           lv = (evaluate_term l env) in 
       	RecCons(lv,c,d)
  | TRecCons (l,v,r) ->
       let d = (evaluate_term r env) and
           c = (evaluate_term v env) and 
           lv = (evaluate_term l env) in 
       	TRecCons(lv,c,d)
  | REmpty e ->
	  let d = (evaluate_term e env) in 
	  (match d with
	      EmptyRec -> Bool true
	    | RecCons _ -> Bool false
	    | TEmptyRec -> Bool true
	    | TRecCons _ -> Bool false
	    | _ ->  failwith "norecord_err")
  | RLabel e ->
	  let d = (evaluate_term e env) in 
	  (match d with
	      RecCons(l,v,r) -> l
	    | TRecCons(l,v,r) -> l
	    | _ ->  failwith "norecord_err")
  | RHead e -> 
	  let d = (evaluate_term e env) in 
	  (match d with
	      RecCons(l,v,r) -> v
	    | TRecCons(l,v,r) -> v
	    | _ ->  failwith "norecord_err")
  | RTail e ->
	  let d = (evaluate_term e env) in 
	  (match d with
	      RecCons(l,v,r) -> r
	    | TRecCons(l,v,r) -> r
	    | _ ->  failwith "norecord_err")
  | TDom e ->
	  let d = (evaluate_term e env) in 
	  (match d with
	    TArrow(d,c) -> d
	   | _ ->  failwith "noarrowt_err")
  | TCodom e ->
	  let d = (evaluate_term e env) in 
	  (match d with
	    TArrow(d,c) -> c
	   | _ ->  failwith "noarrowt_err")
   | TRef e -> 
	let d = (evaluate_term e env)
		in TRef d
   | TRefOf e ->
	let d = (evaluate_term e env) in 
	  (match d with
	    TRef d -> d
	    | _ ->  failwith "noreft_err")

  | Ref e -> let d = (evaluate_term e env) in Loc (ref d)

  | Deref e -> let d = (evaluate_term e env) in 
	  (match d with
	    Loc l -> !l
	    | _ ->  failwith "noref_err")

  | Assign (e1,e2) ->
        let v1= (evaluate_term e1 env) in 
        let v2= (evaluate_term e2 env) in 
	  (match v1 with
	    Loc l -> (l := v2; v2)
	    | _ ->  failwith "noref_err")

  | Equals(a,b,t) ->
	  let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          Bool (a'=b') 

  | And(a,b) ->
	  let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          andPrim a' b'
  | Or(a,b) ->
	  let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          orPrim a' b'
  | Implies(a,b) ->
	  let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          impliesPrim a' b' 
  | Not(a) ->
	  let a' = (evaluate_term a env) in 
          notPrim  a' 

  | EqVal (a,b) -> 
          let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          eqPrim a' b'

  | NumPlus (a,b) -> 
          let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          plusNumPrim a' b'

  | NumMinus (a,b) -> 
          let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          minusNumPrim a' b'

  | NumTimes (a,b) -> 
          let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          timesNumPrim a' b'

  | StrConcat (a,b) -> 
          let a' = (evaluate_term a env) in 
	  let b' = (evaluate_term b env) in 
          strConcatPrim a' b'

  | Tracer (a) -> (evaluate_term a env) 

  (* some cases missing *)
;;



let entails env f =
  let (env',f',extra_decls) = collector env f in
  (*print_endline ("DECLS " ^ extra_decls) ; *)
  (*print_endline "ENTAILS" ;
    print_endline "Env = " ; (List.iter (fun (s,t) -> print_term t) env) ; *)
  let aux (hks,decls,assumps) elem =
    let (hk,decl,assump) = new_ctxt_elem elem in
      (hks^hk,decls@decl,assumps@assump) in
  let (hk,decls,assump) = List.fold_left aux ("",[],[]) env' in
  let form = Vcgen.gen_vc_formula f' in
  let query = Vcgen.prep_query extra_decls hk decls assump form in 
  (*print_endline query ; *)
   Vcgen.make_query query

;;

(*
let entails env t k = true;;
*)

let rec ren_term t env =
  match t with
  |  Bool _ -> t
  |  Num _ -> t
  |  String _ -> t

  | TInt -> t
  | TBool -> t
  | TUnit -> t
  | TString -> t
  | Ttop -> t
  | Tbot -> t

  | KindOK -> t

  | KType -> t
  | KFun -> t
  | KRec -> t
  | KRef -> t
  | KLab -> t

  | KGen k -> KGen (ren_term k env)

  | KRefine (x,k,p) -> let nn = new_id () in KRefine(nn,(ren_term k env),(ren_term p ((x,Id nn)::env))) 

  | KPi (x,u,b) -> (let nn = new_id () in KPi (nn, ren_term u env, ren_term b ((x,Id (nn))::env)))

  |  Ite(c,t,e) ->
	 let v1 = ren_term c env in Ite (v1, ren_term t env, ren_term e env) 

  |  Id s -> (try find s env with _ -> t) 

  |  Lam (x,u,b) -> let nn = new_id () in 
                    Lam (nn, ren_term u env, ren_term b ((x,Id (nn))::env))
  |  TLam (x,u,b) -> let nn = new_id () in ( 
                     TLam (nn, ren_term u env, ren_term b ((x,Id (nn))::env)))

  |  Clos _ -> failwith "nf_term: vclosure unexpected"
  |  TClos _ -> failwith "nf_term: tclosure unexpected"
  |  FClos _ -> failwith "nf_term: tclosure unexpected"

  |  TForall (x,k,b) -> let nn = new_id () in TForall(nn, ren_term k env, ren_term b ((x,Id (nn))::env))

  |  Letval (x,t,d,b) -> let nn = new_id () in
                         let t' = (ren_term t env) in 
                         let d' = (ren_term d env) in 
                         let b' = ren_term b ((x,Id nn)::env) 
                         in Letval (nn,t',d',b')

  |  Lettype (x,t,d,b) -> let nn = new_id () in
                         let t' = (ren_term t env) in 
                         let d' = (ren_term d env) in 
                         let b' = ren_term b ((x,Id nn)::env) 
                         in Lettype (nn,t',d',b')

  |  Letrecval (x,t,d,b) -> let nn = new_id () in
                         let t' = (ren_term t env) in 
                         let env' =  ((x,Id nn)::env) in
                         let d' = ren_term d env' in 
                         let b' = ren_term b env' 
                         in Letrecval (nn,t',d',b')

  |  Letrectype (x,t,d,b) ->  let nn = new_id () in
                         let t' = (ren_term t env) in 
                         let env' =  ((x,Id nn)::env) in
                         let d' = ren_term d env' in 
                         let b' = ren_term b env' 
                         in Letrectype (nn,t',d',b')

  |  App (u,v) -> 
       let v1 = (ren_term u env) and
           v2 = (ren_term v env) in App (v1,v2)

  |  TMap (u,v) -> 
           let v1 = (ren_term u env) and
               v2 = (ren_term v env) in TMap (v1,v2)
       
  |  TArrow (u,v) -> 
       let d = (ren_term u env) and
           c = (ren_term v env) in 
       	TArrow(d,c)
  | TDom e ->
	  let d = (ren_term e env) in TDom d
  | TCodom e ->
	  let d = (ren_term e env) in TCodom d

  |  TRef e -> 
       let d = (ren_term e env) in TRef d
  |  TRefOf e ->
	  let d = (ren_term e env) in TRefOf d

  | EmptyRec -> t
  | TEmptyRec -> t
  | Label s -> t
  | RecCons(l,v,r) -> 
       let d = (ren_term r env) and
           c = (ren_term v env) and
           ll = (ren_term l env) in
       	RecCons(ll,c,d)
  | TRecCons(l,v,r) -> 
       let d = (ren_term r env) and
           c = (ren_term v env) and
           ll = (ren_term l env) in
       	TRecCons(ll,c,d)
  | RLabel e ->
	  let d = (ren_term e env) in RLabel d
  | RHead e -> 
	  let d = (ren_term e env) in RHead d
  | RTail e ->
	  let d = (ren_term e env) in RTail d

  | REmpty e ->
	  let d = (ren_term e env) in REmpty d 

  | Equals(a,b,t) ->
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in 
	  let t' = (ren_term t env) in Equals(a',b',t') 

(* complete this below *)

  | NumMinus(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in NumMinus(a',b') 

  | NumPlus(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in NumPlus(a',b') 

  | NumTimes(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in NumTimes(a',b') 

  | StrConcat (a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in StrConcat(a',b') 

  | And(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in And(a',b') 

  | Or(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in Or(a',b') 

  | Implies(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in Implies(a',b') 

  | Not a -> 
	  let a' = (ren_term a env) in Not(a') 

  | In(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in In(a',b')

  | LabSet(a) -> let a' = (ren_term a env) in LabSet(a')

  | LabApart(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in LabApart(a',b') 

  | LabUnion(a,b) -> 
	  let a' = (ren_term a env) in 
	  let b' = (ren_term b env) in LabUnion(a',b') 

  | RecF(f,t,b) -> 
	  let b' = (ren_term b env) in 
          let t' = (ren_term t env) in RecF(f,t',b') 

  | LSet -> t

(* effects *)

  | Ref e' -> Ref (ren_term e' env)

  | Deref e' -> Deref (ren_term e' env)

  | Assign (e1,e2) ->  Assign (ren_term e1 env,ren_term e2 env)
  
  | _ -> (print_term t;  failwith "REN_MISSING CASE")
;;

let rec ren_termd t env =
  match t with
  |  Bool _ -> t
  |  Num _ -> t
  |  String _ -> t

  | TInt -> t
  | TBool -> t
  | TUnit -> t
  | TString -> t
  | Ttop -> t
  | Tbot -> t

  | KindOK -> t

  | KType -> t
  | KFun -> t
  | KRec -> t
  | KRef -> t
  | KLab -> t

  | KGen k -> KGen (ren_term k env)

  | KRefine (x,k,p) -> let nn = new_id () in KRefine(nn,(ren_term k env),(ren_term p ((x,Id nn)::env))) 

  | KPi (x,u,b) -> (let nn = new_id () in KPi (nn, ren_term u env, ren_term b ((x,Id (nn))::env)))

  |  Ite(c,t,e) -> let v1 = ren_termd c env in Ite (v1, ren_termd t env, ren_termd e env)

  |  Id s  -> t (* (try find s env with _ -> t) *)
 (* do not rename, we are in code part *)

  |  Lam (x,u,b) -> let nn = new_id () in 
                    let b' = ren_term b [x,Id (nn)] in
                    Lam (nn, ren_term u env, ren_termd b' env)
  |  TLam (x,u,b) -> let nn = new_id () in
                     let b' = ren_term b [x,Id (nn)] in
                     TLam (nn, ren_term u env, ren_termd b' env)

  |  Clos _ 
  |  TClos _ 
  |  FClos _ -> failwith "ren_termd: closure unexpected"

  |  TForall (x,k,b) -> let nn = new_id () in 
                        TForall(nn, ren_term k env, ren_term b ((x,Id (nn))::env))

(* HERE *)

  |  Letval (x,t,d,b) -> let nn = new_id () in
                         let t' = ren_term t env in 
                         let b' = ren_term b [x,Id (nn)] in 
                         let d'' = ren_termd d env in 
                         let b'' = ren_termd b' env
                         in Letval (nn,t',d'',b'')

  |  Lettype (x,t,d,b) -> let nn = new_id () in
                         let t' = ren_term t env in 
                         let b' = ren_term b [x,Id (nn)] in 
                         let d'' = ren_term d env in 
                         let b'' = ren_termd b' env
                         in Lettype (nn,t',d'',b'')

  |  Letrecval (x,t,d,b) -> let nn = new_id () in
                         let t' = ren_term t env in 
                         let d' = ren_term d [x,Id (nn)] in 
                         let b' = ren_term b [x,Id (nn)] in 
                         let d'' = ren_termd d' env in 
                         let b'' = ren_termd b' env
                         in Letrecval (nn,t',d'',b'')

  |  Letrectype (x,t,d,b) ->  let nn = new_id () in
                         let t' = ren_term t env in 
                         let d' = ren_term d [x,Id (nn)] in 
                         let b' = ren_term b [x,Id (nn)] in 
                         let d'' = ren_term d' env in 
                         let b'' = ren_termd b' env
                         in Letrectype (nn,t',d'',b'')

  |  App (u,v) -> 
       let v1 = (ren_termd u env) and
           v2 = (ren_termd v env) in App (v1,v2)

  |  TMap (u,v) -> 
           let v1 = (ren_termd u env) and
               v2 = (ren_termd v env) in TMap (v1,v2)
       
  |  TArrow (u,v) -> 
       let d = (ren_termd u env) and
           c = (ren_termd v env) in 
       	TArrow(d,c)
  | TDom e ->
	  let d = (ren_termd e env) in TDom d
  | TCodom e ->
	  let d = (ren_termd e env) in TCodom d

  |  TRef e -> 
       let d = (ren_termd e env) in TRef d
  |  TRefOf e ->
	  let d = (ren_termd e env) in TRefOf d

  | EmptyRec -> t
  | TEmptyRec -> t
  | Label s -> t
  | RecCons(l,v,r) -> 
       let d = (ren_termd r env) and
           c = (ren_termd v env) and
           ll = (ren_termd l env) in
       	RecCons(ll,c,d)
  | TRecCons(l,v,r) -> 
       let d = (ren_termd r env) and
           c = (ren_termd v env) and
           ll = (ren_termd l env) in
       	TRecCons(ll,c,d)
  | RLabel e ->
	  let d = (ren_termd e env) in RLabel d
  | RHead e -> 
	  let d = (ren_termd e env) in RHead d
  | RTail e ->
	  let d = (ren_termd e env) in RTail d

  | REmpty e ->
	  let d = (ren_termd e env) in REmpty d 

  | Equals(a,b,t) ->
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in 
	  let t' = (ren_term t env) in Equals(a',b',t') 

(* complete this below *)

  | NumMinus(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in NumMinus(a',b') 

  | NumPlus(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in NumPlus(a',b') 

  | NumTimes(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in NumTimes(a',b') 

  | StrConcat (a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in StrConcat(a',b') 

  | And(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in And(a',b') 

  | Or(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in Or(a',b') 

  | Implies(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in Implies(a',b') 

  | Not a -> 
	  let a' = (ren_termd a env) in Not(a') 

  | In(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in In(a',b')

  | LabSet(a) -> let a' = (ren_termd a env) in LabSet(a')

  | LabApart(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in LabApart(a',b') 

  | LabUnion(a,b) -> 
	  let a' = (ren_termd a env) in 
	  let b' = (ren_termd b env) in LabUnion(a',b') 

  | RecF(f,t,b) -> 
	  let b' = (ren_termd b env) in 
          let t' = (ren_term t env) in RecF(f,t',b') 

  | LSet -> t

(* effects *)

  | Ref e' -> Ref (ren_termd e' env)

  | Deref e' -> Deref (ren_termd e' env)

  | Assign (e1,e2) ->  Assign (ren_termd e1 env,ren_termd e2 env)
  
  | _ -> (print_term t;  failwith "REND_MISSING CASE")
;;

let struct_inv t  = ((TRecCons ((RLabel t), (RHead t ), (RTail t))));;

let rec struct_cons t e  =
  match t with
  | TEmptyRec | TRecCons _ | TArrow _ |  TInt | TBool -> (true, t)
  | Id _ -> (entails e (Equals(t,struct_inv t,KType)), struct_inv t)  
  | _ -> (false,Tbot)
;;

let rec nf_term t et env =
  match t with
  |  Bool _ -> t
  |  Num _ -> t
  |  String _ -> t

  | TInt -> t
  | TBool -> t
  | TUnit -> t
  | TString -> t
  | Ttop -> t
  | Tbot -> t

  | KindOK -> t

  | KType -> t
  | KFun -> t
  | KRec -> t
  | KRef -> t
  | KLab -> t

  | KGen k -> KGen (nf_term k et env)

  | KRefine (x,k,p) -> let nn = new_id () in KRefine(nn,(nf_term k et env),(nf_term p et ((x,Id nn)::env))) 

  | KPi (x,u,b) -> let nn = new_id () in KPi (nn, nf_term u et env, nf_term b et  ((x,Id (nn))::env))

  |  TForall (x,k,b) -> 
       let nn = new_id () in (
         let k'= nf_term k et env in ( 
         let b'= nf_term b et ((nn,Id nn)::(x,Id (nn))::env) in (
          TForall(nn, k',b')
       )))

  |  Ite(c,tb,eb) ->
	 let v1 = nf_term c et env in
       (
         match v1 with
            Bool true -> nf_term tb et env
	 |  Bool false -> nf_term eb et env
 	 |  _ -> (if entails et v1 then (nf_term tb et env) else
                  if entails et (Not v1) then nf_term eb et env else
                  Ite (v1, nf_term tb et env, nf_term eb et env))) 

  |  Id s -> find s env 

  |  Lam (x,u,b) -> let nn = new_id () in 
                    Lam (nn, nf_term u et env, nf_term b et ((x,Id (nn))::env)) 

  |  TLam (x,u,b) -> let nn = new_id () in
                     TLam (nn, nf_term u et env, nf_term b et ((x,Id (nn))::env)) 

  |  RecF (f,t,b) ->RecF (f, nf_term t et env, nf_term b et ((f,Id f)::env)) 

  |  Clos _ -> failwith "nf_term: vclosure unexpected"

  |  TClos _ -> failwith "nf_term: tclosure unexpected"

  |  FClos _ -> failwith "nf_term: tclosure unexpected"

  |  Letval (x,t,d,b) -> 
       let v1 = (nf_term d et env) in nf_term b et ((x,v1)::env) 

  |  Lettype (x,t,d,b) -> 
       let v1 = (nf_term d et env) in nf_term b et ((x,v1)::env)

  |  Letrecval (x,t,d,b) -> 
       let t1 = (nf_term t et env) in
       let v1 = (nf_term d et ((x,Id(x))::env)) in
       let v2 = (nf_term b et ((x,Id(x))::env)) in
       Letrecval (x,t1,v1,v2) 

  |  Letrectype (x,t,d,b) -> 
       let t1 = (nf_term t et env) in
       let v1 = (nf_term d et ((x,Id(x))::env)) in
       let v2 = (nf_term b et ((x,RecF(x,t1,v1))::env)) in v2

  |  App (u,v) -> ( 
       let v1 = (nf_term u et env) and
           v2 = (nf_term v et env) in 
       (
        match v1 with 
          Lam(x,_,b) -> let nn = new_id() in (
                        let b' = nf_term b et ((x,Id nn)::env) in 
                        nf_term b' et ((nn,v2)::env))
        | TLam(x,_,b)-> let nn = new_id() in (
                        let b' = nf_term b et ((x,Id nn)::env) in
                        nf_term b' et ((nn,v2)::env))
 (* print_term v2; print_string (" <<-- ARG FNAME = "^f^"  \n\n"); *)
        | RecF (f,_,z) -> (
                           match z with 
                           | TLam (x,_,b) -> (  
                                                let (fl,ste) = struct_cons v2 et in
                                                if fl then (let b' = ren_term b [ (f, v1)  ] in nf_term b' et ((x,ste)::env)) else App(v1,v2))

                           | _ -> failwith "No recursive value") 
        | _ -> App (v1,v2)
       ))

  |  TMap (u,v) -> 
           let v1 = (nf_term u et env) and
               v2 = (nf_term v et env) in 
           (match v1 with 
              | TForall (x,_,b) -> nf_term b et ((x,v2)::env)
              | FClos (x,b,e) -> nf_term b et ((x,v2)::e)
              | _ -> TMap (v1,v2)
       )
  |  TArrow (u,v) -> 
       let d = (nf_term u et env) and
           c = (nf_term v et env) in 
       	TArrow(d,c)
  | TDom e ->
	  let d = (nf_term e et env) in 
	  (match d with
	    TArrow(d,c) -> d
	   | _ ->  TDom d)
  | TCodom e ->
	  let d = (nf_term e et env) in 
	  (match d with
	    TArrow(d,c) -> c
	   | _ ->  TCodom d)

  |  TRef e -> 
       let d = (nf_term e et env)
		in TRef d
  |  TRefOf e ->
	  let d = (nf_term e et env) in 
	  (match d with
	    TRef r -> r
	   | _ ->  TRefOf d)

  | EmptyRec -> t
  | TEmptyRec -> t
  | Label s -> t
  | RecCons(l,v,r) -> 
       let d = (nf_term r et env) and
           c = (nf_term v et env) and
           ll = (nf_term l et env) in
       	RecCons(ll,c,d)
  | TRecCons(l,v,r) -> 
       let d = (nf_term r et env) and
           c = (nf_term v et env) and
           ll = (nf_term l et env) in
       	TRecCons(ll,c,d)
  | RLabel e ->
	  let d = (nf_term e et env) in 
	  (match d with
	      RecCons(l,v,r) -> l
         | TRecCons(l,v,r) -> l
	    | _ ->  RLabel d)
  | RHead e -> 
	  let d = (nf_term e et env) in 
	  (match d with
	      RecCons(l,v,r) -> v
         | TRecCons(l,v,r) -> v
	    | _ -> RHead d)
  | RTail e ->
	  let d = (nf_term e et env) in 
	  (match d with
	      RecCons(l,v,r) -> r
         | TRecCons(l,v,r) -> r
	    | _ -> RTail d)
  | REmpty e ->
	  let d = (nf_term e et env) in 
	  (match d with
	      EmptyRec -> Bool true
	    | RecCons _ -> Bool false
	    | TEmptyRec -> Bool true
	    | TRecCons _ -> Bool false
	    | _ -> REmpty d )
  | Equals(a,b,t) ->
	  let a' = (nf_term a et env) in  (* incomplete check, lacks entails *)
	  let b' = (nf_term b et env) in 
	  let t' = (nf_term t et env) in (
          if (a'=b') then (Bool true) 
          else Equals(a',b',t') )

(* complete this below *)

  | NumMinus(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Num b1, Num b2 -> minusNumPrim a' b'
          | _,_ -> NumMinus(a',b') )

  | NumPlus(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Num b1, Num b2 -> plusNumPrim a' b'
          | _,_ -> NumPlus(a',b') )

  | NumTimes(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Num b1, Num b2 -> timesNumPrim a' b'
          | _,_ -> NumTimes(a',b') )

  | StrConcat (a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | String b1, String b2 -> strConcatPrim a' b'
          | _,_ -> StrConcat(a',b') )

  | And(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Bool b1, Bool b2 -> andPrim a' b'
          | _,_ -> And(a',b') )

  | Or(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Bool b1, Bool b2 -> orPrim a' b'
          | _,_ -> Or(a',b') )

  | Implies(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in 
          match a',b' with 
          | Bool b1, Bool b2 -> impliesPrim a' b'
          | _,_ -> Implies(a',b') )

  | Not a -> (
	  let a' = (nf_term a et env) in 
          match a' with 
          | Bool b -> notPrim a'
          | _ -> Not(a') )

  | In(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in In(a',b'))

  | LabSet(a) -> let a' = (nf_term a et env) in LabSet(a')

  | LabApart(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in LabApart(a',b') )

  | LabUnion(a,b) -> (
	  let a' = (nf_term a et env) in 
	  let b' = (nf_term b et env) in LabUnion(a',b') )

  | LSet -> t

(* effects *)

  | Ref _   | Deref _   | Assign _ -> failwith "effect not allowed"

;;

let rec unrefine t =
  match t with
  | KRefine (id,k,p) -> unrefine k
  | _-> t
;;

let rec hnf f =
  match f with
  | Id x -> true
  | App (g,a) -> hnf g
  | _ -> false;;

let rec subtype (et:termenv) (e:termenv) (t1:term) (t2:term) = (* types assumed NF *)

   if !Vcgen.debug then (print_term t1; print_string " !! - match\n"; 
       print_term t2; print_string " !! - match with \n";
       print_env e; print_string " !! - with ENVV \n");

   match t1, t2 with

  |   KindOK, KindOK -> true 

  |   KType, KType -> true 

  |   Ttop, Ttop -> true 

  |   Id _, Ttop -> true (* Id must be type var !! *)
  |   Id x, Id y -> x == y

  |   KFun, KType -> true 
  |   KFun, KFun -> true 

  |   KRef, KType -> true 
  |   KRef, KRef -> true 

  |   KLab, KType -> true 
  |   KLab, KLab -> true 

  |   KRec, KType -> true 
  |   KRec, KRec -> true 
 
  |   KRefine _, KType -> true 

  |   KGen _, KType -> true
  |   KGen k,  KGen k' -> subtype et e k' k

  |   TInt, Ttop -> true
  |   TInt, TInt -> true

  |   TString, Ttop -> true
  |   TString, TString -> true

  |   TBool, Ttop -> true
  |   TBool, TBool -> true

  |   TRef _, Ttop -> true
  |   TRef t, TRef t' -> subtype et e t t'

  |   TEmptyRec, TEmptyRec -> true
  |   TEmptyRec, Ttop -> true

  |   Ite(c,t,f), Ite(c',t',f') -> (c=c') && subtype et e t t' && subtype et e f f'

  |   TRecCons(l,t,r), Ttop -> true
  |   t1, TRecCons(l,t,r) -> (
         let td = nf_term  (RHead t1) et e and tc = nf_term  (RTail t1) et e in (
         (subtype et e td (nf_term t et e)) && (subtype et e tc (nf_term r et e)))
         )

  |   TArrow (s,t), Ttop -> true
  |   t1, TArrow (s,t) -> (
         let td = nf_term (TDom t1) et e and tc = nf_term  (TCodom t1) et e in (
         (subtype et e td (nf_term s et e)) && (subtype et e tc (nf_term t et e)))
         )

  |   TForall (x,s,t), Ttop -> true
  |   TForall (x,s,t) , TForall (x',s',t') -> ( 
             let c = new_id () in (
             let et' = (c,s)::et in
             let e' = (c,Id c)::e in
             let rn  = [(c,Id c);(x,Id c);(x',Id c)] in
             let v  = ren_term t rn in
             let v' = ren_term t' rn in
             (subtype et e s s') && (subtype et' e' v v'))
         )

  |   KPi (x,s,t), KType -> true
  |   KPi (x,s,t) , KPi (x',s',t') -> ( 
             let c = new_id () in 
             let et' = (c,s)::et in
             let e' = (c,Id c)::e in
             let rn  = [ (x,Id c);(x',Id c)] in
             let v  = ren_term t rn in
             let v' = ren_term t' rn in
             (subtype et e s s') && (subtype et' e' v v') 
         )

  |   KRefine(x1,k1,p1), KRefine(x2,k2,p2) -> (
              (* print_string "KRefine match !!" ; print_term  (KRefine(x1,k1,p1)); print_string " - "; print_term ( KRefine(x2,k2,p2)) ;print_string "\n!";*)
              let nn = (new_id ()) in
              let et' = (nn,k1)::et in 
              let rn = [(x2,Id nn);(x1,Id nn)]  in
              let pnf =  ren_term (Implies(p1,p2)) rn in
              ((subtype et e k1 k2) && entails et' pnf))  

  |   KRefine(_,k1,_), _ -> subtype et e k1 t2

  |   RHead(t), Ttop -> subtype et e t Ttop   

 (* |   RHead t, RHead t' -> subtype et e t t'  *)

  |   RTail(_), Ttop -> true   (*HNF *)
 
(* |   RTail t, RTail t' -> subtype et e t t'  *)

  |   TDom(_), Ttop -> true   (*HNF *)
 (* |   TDom t, TDom t' -> subtype et e t t'  *)

  |   TCodom t, TCodom t' -> subtype et e t t'  

  |   TCodom(_), Ttop -> true   (*HNF *)

  |   RLabel(_), KType -> true (* HNF *)
  |   RLabel(_), KLab -> true (* HNF *)

  |   TRefOf _, Ttop -> true
  |   TRefOf t, TRefOf t' -> subtype et e t t'

  |   LSet, LSet -> true

  |   TLam (x,s,t) , TLam (x',s',t') -> (
             let c = new_id () in 
             let et' = (c,s)::et in
             let e' = (c,Id c)::e in
             let rn  = [ (x,Id c);(x',Id c)] in
             let v  = ren_term t rn in
             let v' = ren_term t' rn in
             (subtype et e s s') && (subtype et' e' v v') 
         )

  |   App (f,a), Ttop -> true

  |   App (f,a), App(f',a') -> (subtype et e f f' && subtype et e a a' )

  |   RecF (f,t,b), RecF(f',t',b') -> (f = f')

  |   t1, t2  -> (entails et (Equals(t1,t2,KType)))
;;

	
let validate_type e ee t1 t2 = 
 let t1' = nf_term t1 e ee in (* should not be needed, if invariants are preserved *)
 let t2' = nf_term t2 e ee in
  (if subtype e ee t1' t2' then t1' 
      else (if !Vcgen.debug then (print_term t1'; print_string "\n\n - \n\n"; print_term t2'; print_string " Xvalidate\n"); 
                failwith "term does not match expected type or kind"));;

let join_kind e ee k1 k2 = 
        if (subtype e ee k1 k2) then k2 
        else if (subtype e ee k2 k1) then k1
        else KType

(* invariant: typecheck returns nfs *)

let isunsound env = 
  if !Vcgen.debug then (print_string "isunsound ENV: "; print_env env);
   entails env (Bool false) ;; 

let assertKPI t = 
  match t with
  |  KPi _ -> t 
  |  _ -> (print_string "\n"; print_term t; failwith "Pi Type Expected")
;;

let assertTForall t = 
  match t with
  |  TForall _ -> t 
  |  _ -> (print_string "\n"; print_term t; failwith "All Type Expected")
;;

let rec inductive2 f ih = 
  match ih with 
  | [] -> None
  | (x,arg,t)::r -> if x = f then Some (x,arg,t) else inductive2 f r
;;

let rec less_than t1 t2  = (* incomplete *)
(t1=RTail(t2))
;;

let rec headlst t lst = 
  match t with 
  | App (f,a) -> headlst f (a::lst)
  | h -> h::lst
;;

let rec applst t lst =
   match lst with
   | [] -> t 
   | h::r -> applst (App(t,h)) r
;;

let rec typecheck e ee t1 t2 t2k ff ih  =

        if !Vcgen.debug then
        (print_term t1; print_string " !! - type \n"; 
        print_term t2; print_string " !! - expect type \n";
        print_term t2k; print_string " !! - expect kind \n";
        print_string " ff: "; print_term (Bool ff) ; print_string " \n";
        print_string " ENVT: "; print_env e ; print_string " \n";

        print_string " ENVV: "; print_env ee; print_string " \n";
        print_string " IH: "; print_ih ih ; print_string " GO TC\n" );

        if ff then t2 else

        match t2 with (* conservative inversions first on expected type position *)

        |  App(f,s) ->  ( 

                        (* print_term t1; print_string " : \n"; print_term (App(f,s)); print_string "  <- TYPE EXPECTED APP\n"; *)

                        let head::rest = headlst t2  [] in (
                        match head with
                        |  Ite(c,tb,eb) -> (typecheck e ee t1 (Ite (c, (applst tb rest), (applst eb rest))) t2k ff ih)
                        |  RecF(f,t,b) ->  ( match (inductive2 f ih) with 
                                             | Some (g,arg,t1') -> ( if (less_than (List.hd rest) arg) then 
                                                                     (typecheckleft e ee t1 t2 t2k ff ih)
                                                                    else failwith "invalid induction hyp")
                                             | None -> ( 
                                                         let arg = (List.hd rest) in
                                                         let (fl,ste) = struct_cons arg e in 
                                                         if fl then ( (* unfold *)
                                                         let b' = nf_term b ((f,t)::e) ((f,head)::ee) in (
                                                         let b'' = applst b' rest in 
                                                         let nfb'' = (nf_term b'' e ee) in ( 
                                                          typecheck e ee t1 nfb'' t2k ff ((f,(List.hd rest),t1)::ih))
                                                         ))
                                                         else 
                                                          typecheckleft e ee t1 t2 t2k ff ih
                                                        )
                                           )
                        | Id f ->  (typecheck e ee t1 (nf_term t2 e ee) t2k ff ih )
                        | TLam (x,tx,b) ->  ( let nf2 = (nf_term t2 e ee) in 
                                             typecheck e ee t1 nf2 t2k ff ih )
                        | _ ->  ( print_term head; print_string "\n";failwith "invalid head at expected type" )
                        ))
  

        |  Ite(c,tb,eb) -> (
                           let _ = typecheckrefinement e ee c TBool ff in 
                           let cnf = (nf_term c e ee) in (
                             (* print_string "\n\n IF condition ="; print_term cnf; print_string "\n\n"; *)
                           if cnf = Bool true then 
                               ( let _ = typecheck e ee t1 tb t2k ff ih in ( Ite(cnf,tb,eb) )) else
                           if cnf = Bool false then 
                               ( let _ = typecheck e ee t1 eb t2k ff ih in ( Ite(cnf,tb,eb) )) 
                           else let idx = new_id () in ( (* print_string "\n\nITE try both branches\n\n\n"; *)
                                 let ethen = ((idx, KRefine(idx,KType,    cnf))::e) in
                                 let eelse = ((idx, KRefine(idx,KType,Not cnf))::e) in
                                 let _ = typecheck ethen ee t1 tb t2k (isunsound ethen) ih in (  
                                 let _ = typecheck eelse ee t1 eb t2k (isunsound eelse) ih in (   
                                   Ite(cnf,tb,eb))))))

                                 (* validate_type e ee (join_kind e ee tk ek) t2) *)

 (* print_string "\n\nTYpe THEN OK\n"; *)
 (* print_string "\n\nTYpe ELSE OK\n"; *)

        |  KRefine(x,k,p) -> ( let kk = typecheck e ee t1 k KindOK ff ih in  
                     let v1 = t1 in (  
                     let ps = (nf_term p e ((x,v1)::ee)) in
                     let b = entails e ps in ( 
                     if b then validate_type e ee kk k 
                     else (print_term p; (* print_string "\n"; print_env e; *) failwith "\nVC gen fail" ))))

        |  _ -> typecheckleft e ee t1 t2 t2k ff ih

and typecheckleft e ee t1 t2 t2k ff ih = 

        if !Vcgen.debug then
        (print_term t1; print_string " !! - type \n"; 
        print_term t2; print_string " !! - expect type \n";
        print_term t2k; print_string " !! - expect kind \n";
        print_string " ff: "; print_term (Bool ff) ; print_string " \n";
        print_string " ENVT: "; print_env e ; print_string " \n";

        print_string " ENVV: "; print_env ee; print_string " \n";
        print_string " IH: "; print_ih ih ; print_string " GO TC LEFT \n" );

        ( 
	match t1 with
                Id x -> validate_type e ee (find x e) t2

 	|	Num _ -> validate_type e ee TInt t2
 	|	Bool _ -> validate_type e ee TBool t2
 	|	String _ -> validate_type e ee TString t2

 	|	KFun -> validate_type e ee KindOK t2 
 	|	KLab -> validate_type e ee KindOK t2 
 	|	KRef -> validate_type e ee KindOK t2 
 	|	KType -> validate_type e ee KindOK t2 
        |       KRec -> validate_type e ee KindOK t2

 	|	KGen u -> let uk = typecheck e ee u KindOK KindOK ff ih 
                          in validate_type e ee uk t2 

 	|	KRefine (x,t,u) -> ( 
                   let KindOK = typecheck e ee t KindOK KindOK ff ih in 
                   let     _ = typecheckrefinement ((x,t)::e) ((x,Id x)::ee) u TBool 
                   in ( validate_type e ee KindOK t2))

 	|	TInt -> validate_type e ee KType t2
	|	TBool -> validate_type e ee KType t2
	|	TString -> validate_type e ee KType t2
	|	TUnit -> validate_type e ee KType t2
	|	Ttop -> validate_type e ee KType t2
	|	Tbot -> validate_type e ee KType t2

        |       Ite(c,a,b) -> (  (*  print_string "\n\n\n"; print_term a; print_string "  - ";
                                     print_term b; print_string " ET= ";  print_term t2; print_string " ITELSE ==>> \n\n\n";  count_debug() ;  *)
			let idx = new_id () in
			let _ = typecheckrefinement e ee c TBool  in ( 
                        let et = ((idx,KRefine(idx,KType,c))::e) in
			let ak = typecheck et ee a t2 t2k (isunsound et) ih in ( 
                        (* print_string "\n\nTHEN branch OK\n\n\n"; *)
                        let ef = ((idx,KRefine(idx,KType,Not c))::e) in 
			let bk = typecheck ef ee b t2 t2k (isunsound ef) ih in (   
                        (* print_string "\n\nBOTH branches OK\n\n\n"; print_term ak; print_string "\n"; print_term bk; *)
                         let ak' = (nf_term ak et ee) in
                           let bk' = (nf_term bk ef ee) in (
                             if ak' = bk' then (ak) (* use equality check *)
                             else (
                           let rt = validate_type e ee (join_kind e ee ak' bk') t2 in (rt)))))
                          ))
	| 	TArrow (d,c) -> 
			let _ = typecheck e ee d KType KindOK ff ih
			and _ = typecheck e ee c KType KindOK ff ih
			in validate_type e ee KFun t2
	| 	TRef r -> 
			let _ = typecheck e ee r KType KindOK ff ih
		        in validate_type e ee KRef t2

        |       Letval (x,t,d,b) -> ( (* print_term t; print_string " \n\nLETVAL\n\n"; *)
                        let tnf = nf_term t e ee in (
			let ke = typecheck e ee tnf KType KindOK ff ih in (
			let _ = typecheck e ee d t ke ff ih in (* check if t2t' is kind *)
                        let e' =  ((x,tnf)::e) in
		        typecheck e' ((x,Id x)::ee) b t2 t2k (isunsound e')) ih))

        |       Lettype (x,t,d,b) -> (
                        let nn = new_id () in
			let tk = typecheck e ee t KindOK KindOK ff ih in
                        let rho = [(x,Id nn)] in
                        let b' = ren_term b rho in 
                        let e'  = (nn,t)::e in 
                        let ee' = (nn,d)::ee in ( (* Id nn *)
			let _ = typecheck e ee d t tk ff ih in  
                        (  
		        typecheck e' ee' (ren_termd b' [nn,d]) t2 t2k (isunsound e') ih) ) )

        |       Letrecval (x,t,d,b) -> 
                        let tnf = (nf_term t e ee) in
			let tk = typecheck e ee tnf KType KindOK ff ih in
                        let e'  = (x,tnf)::e in 
                        let ee' = (x,Id x)::ee in 
			let _ = typecheck e' ee' d tnf tk ff  ih in ( 
		        typecheck e' ee' b t2 t2k (isunsound e') ih)

        |       Letrectype (x,t,d,b) -> (
                        let nn = new_id () in
			let tk = typecheck e ee t KindOK KindOK ff ih in
                        let rho = [(x,Id nn)] in
                        let d' = ren_term d rho in 
                        let b' = ren_term b rho in 
                        let e'  = (nn,t)::e in 
                        let ee' = (nn,Id nn)::ee in (
			let _ = typecheck e' ee' d' t tk ff ih in  (* rename G -> newid add newid -> ren_term d G/newid to ee *)
                        let _ = (nn,(RecF (nn,t,d')))::ee in (  (* we now understand nn as free in RecF important to RecF because of the specific app disciplie *)
		        typecheck e' ee' (ren_term b' [nn,(RecF (nn,t,d'))]) t2 t2k (isunsound e') ih) ) )

        |       TLam (s,k,u) -> (  (* normalize k, ke, etc, as types may reduce *) 
                        match t2 with 
                        | TForall(se,ke,te) -> 
                            (let nn = new_id() in 
                             let _ = validate_type e ee k ke in 
                             let e' = (nn,ke)::e in 
                             let ee' = (nn,Id nn)::ee in 
                             let rn =  [(s,Id nn);(se,Id nn)] in 
                             let u' =  (ren_term u rn) in
                             let te' =  (ren_term te rn) in
                             let _ = typecheck e' ee' u' te' KType (isunsound e') ih in 
                             validate_type e ee (TForall(nn,ke,te')) t2)
                        | KPi(es,ke,ked) -> ( 
                             let nn = new_id() in 
                             let _ = validate_type e ee k ke in
                             let rn = [(s,Id nn);(es,Id nn)] in (
                             let u' = ren_term u rn in 
                             let ked' = ren_term ked rn in ( (* print_term u'; print_string " <- u' \n\n";  print_term ked'; print_string " <- ked' \n\n"; *)
                             let e' = (nn,k)::e in 
                             let ee' = (nn,Id nn)::ee in 
                             let _ = typecheck e' ee' u' ked' KindOK  (isunsound e') ih in 
                             ( (* print_term (KPi (nn, ke, ked' )); print_string "-"; print_term t2; print_string "2 VAL \n";*)
                              validate_type e ee (KPi (nn, ke, ked' )) t2 )
                             )))
                        | _ -> 
                        (
			let _ = typecheck e ee k KindOK KindOK ff ih
                        in if subtype e ee t2 Ttop then 
                            (  let e' = (s,k)::e in
                               let ee' = ((s,Id s)::ee) in 
                              let t = typecheck e' ee' u t2 KType (isunsound e') ih in    (* eg., fun s::k -> 2 : forall X:k.int *)
                              validate_type e ee (TForall(s, k, t)) t2)
                           else if subtype e ee t2 KType then 
                            (                        
                              let e' = (s,k)::e in
                              let ee' = ((s,Id s)::ee) in 
                              let t = typecheck e' ee' u t2 KindOK (isunsound e') ih  (* eg., fun s::k -> int :: Pi s:k.Type  *)
                              in ( 
                                   validate_type e ee (KPi(s,k,t))  t2)
                            ) else ( print_term t2; print_string "\n"; failwith "unreachable!")
                         )
                         )
        |       Lam (s,t,u) -> (
                        match t2 with
                        |  TArrow (ed,ec) -> ( (* print_string "\n\nLAM EXPECT = "; print_term t2; print_string "\n\n"; *)
                               let _ = validate_type e ee ed t in
                               let d = nf_term ed e ee in 
                               let _ = typecheck ((s,d)::e) ((s,Id s)::ee) u ec KType ff ih
                               in validate_type e ee (TArrow(ed,(nf_term ec e ee))) t2
                             )
                        |  _ -> ( (* can only be KType *)
			let _ = typecheck e ee t KType KindOK ff ih in
                        let d = nf_term t e ee in 
                        let cod = typecheck ((s,d)::e) ((s,Id s)::ee) u Ttop KType ff ih
                        in (
                        validate_type e ee (TArrow(d,(nf_term cod e ee))) t2
                        )))

        |       TForall(s,k,t) -> 
			let _ = typecheck e ee k KindOK KindOK ff ih in
                        let e' = ((s,k)::e) in
			let _ = typecheck e'  ((s,Id s)::ee) t KType KindOK (isunsound e') ih
			in validate_type e ee (KGen (nf_term k e' ee)) t2

        |       KPi(s,k,t) -> 
			let _ = typecheck e ee k KindOK KindOK ff ih in
                        let e' = ((s,k)::e) in
			let _ = typecheck e' ((s,Id s)::ee) t KindOK KindOK (isunsound e') ih
			in validate_type e ee KindOK t2

         |       App (s,t) -> typecheckapp e ee t1 t2 t2k ff ih   

                                         
         | Label l  ->  validate_type e ee KLab t2

         | EmptyRec  ->  TEmptyRec

         | RecCons(l,v,r)  -> 
                        let _ = typecheck e ee l KLab KindOK ff ih in
                        let vt = typecheck e ee v Ttop KType ff ih in
                        let rt = typecheck e ee r Ttop KType ff ih in 
                        let _ = typecheck e ee rt KRec KType ff ih in 
                        validate_type e ee (TRecCons((nf_term l e ee),(nf_term vt e ee),(nf_term rt e ee))) t2

         | RHead(a)  ->  (let nn = new_id () in ( 
                              if t2k = KindOK then (
                                let _ = typecheck e ee a (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in  
                                validate_type e ee KType t2 )
                              else if t2k = KType then (
                                 let at = typecheck e ee a Ttop KType ff ih in 
                                  let _ = typecheck e ee at (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in 
                                 let ht  = nf_term (RHead at) e ee in ( 
                                   validate_type e ee ht t2 ))
                              else failwith "RHead: illegal expected kind"))

         | RTail(a)  ->  (let nn = new_id () in (  
                             if t2k = KindOK then (
                         let _ = typecheck e ee a (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in 
                         validate_type e ee KRec t2 )
                              else if t2k = KType then ( 
                         let at = typecheck e ee a Ttop KType ff ih in ( 
                         let _ = typecheck e ee at (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in 
                         let tt  = nf_term (RTail at) e ee in (
                         validate_type e ee tt t2 )))
                                else failwith "RTail: illegal expected kind"))
(*               
         | RLabel(a)  ->  (let nn = new_id () in ( 
                              if t2k = KindOK then (
                                let _ = typecheck e ee a (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in  
                                validate_type e ee KLab t2 )
                              else if t2k = KType then (
                                 let at = typecheck e ee a Ttop KType ff ih in 
                                  let _ = typecheck e ee at (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in 
                                   validate_type e ee KLab t2 )
                              else failwith "RLabel: illegal expected kind"))
*)
         | RLabel(a)  ->
           (let nn = new_id () in ( 
                              try (
                                let _ = typecheck e ee a (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in  
                                validate_type e ee KLab t2 )
                              with _ -> (
                                 let at = typecheck e ee a Ttop KType ff ih in 
                                  let _ = typecheck e ee at (KRefine(nn,KRec,(Not(REmpty(Id nn))))) KindOK ff ih in 
                                   validate_type e ee KLab t2 )))

         | REmpty(a)  -> (try
                         let _ = typecheck e ee a KType KindOK ff ih
                         in validate_type e ee TBool t2
                         with _ -> (
                            let at = typecheck e ee a Ttop KType ff ih
                            in let _ = typecheck e ee at KRec KindOK ff ih
                            in validate_type e ee TBool t2 ))

         | TEmptyRec  ->  KRec

         | TRecCons(l,v,r)  -> 
                        let _ = typecheck e ee l KLab KindOK ff ih in
                        let _ = typecheck e ee v KType KindOK ff ih in
                        let x = new_id () in
                        let rt = typecheck e ee r (KRefine(x,KRec,(Not (In (l, (LabSet (Id x))))))) KindOK ff ih in 
                        let _ = typecheck e ee rt KindOK KindOK ff ih in 
                        validate_type e ee KRec t2

         | TDom(a)  ->   let _ = typecheck e ee a KFun KindOK ff ih in  
                          validate_type e ee KType t2

         | TCodom(a)  ->  
                        let _ = typecheck e ee a KFun KindOK ff ih in  
                          validate_type e ee KType t2

         | TRefOf a   ->
                         let _ = typecheck e ee a KRef KindOK ff ih in  
                         let ct = typecheck e ee (nf_term (TRefOf a) e ee) KType KindOK ff ih in
                          validate_type e ee ct t2

         | Ref a   ->
                         let at = typecheck e ee a Ttop KType ff ih in  
                         validate_type e ee (TRef at) t2

         | Deref a   ->
                         let rt = typecheck e ee a Ttop KType ff ih in  
                         let _ = typecheck e ee rt KRef KindOK ff ih in
                         validate_type e ee (nf_term (TRefOf rt) e ee) t2

         | Assign (a,b)   ->
                         let rt = typecheck e ee a Ttop KType ff ih in  
                         let _ = typecheck e ee rt KRef KindOK ff ih in
                         let vt = (nf_term (TRefOf rt) e ee) in
                         let _ = typecheck e ee b vt KType ff ih in  
                         validate_type e ee vt t2

         | Not a ->  let _ = typecheck e ee a TBool KType ff ih in 
                          validate_type e ee TBool t2

         | And (a,b) ->  let _ = typecheck e ee a TBool KType ff ih in 
                              let _ = typecheck e ee b TBool KType ff ih in
                              validate_type e ee TBool t2

         | Or (a,b) ->  let _ = typecheck e ee a TBool KType ff ih in 
                             let _ = typecheck e ee b TBool KType ff ih in
                             validate_type e ee TBool t2
         | Implies (a,b) ->
                             let _ = typecheck e ee a TBool KType ff ih in 
                             let _ = typecheck e ee b TBool KType ff ih in
                             validate_type e ee TBool t2
         | Equals (a, b, c) ->
                             let tk = if subtype e ee c Ttop then c else KType in
                             let at = typecheck e ee a c tk ff ih in 
                             let bt = typecheck e ee b c tk ff ih in
                             let ak' = (nf_term at e ee) in
                             let bk' = (nf_term bt e ee) in 
                             if ak' == bk' then validate_type e ee TBool t2 else failwith "Kind mismatch in (t == u)"

         | NumPlus (a,b) ->  let _ = typecheck e ee a TInt KType ff ih in 
                              let _ = typecheck e ee b TInt KType ff ih in
                              validate_type e ee TInt t2

         | NumTimes (a,b) ->  let _ = typecheck e ee a TInt KType ff ih in 
                              let _ = typecheck e ee b TInt KType ff ih in
                              validate_type e ee TInt t2

         | StrConcat (a,b) ->  let _ = typecheck e ee a TString KType ff ih in 
                              let _ = typecheck e ee b TString KType ff ih in
                              validate_type e ee TString t2

         | RecF (f, ty, b) ->  ty


         )

and typecheckapp e ee t1 t2 t2k ff ih = let arty = (
                   match t1 with
                   App(s,t) -> (
                   (* print_string "\n\nAPP EXPECTED TYPE="; print_term t1; print_string " : "; print_term t2; print_string " :: "; print_term t2k; print_string " ** \n"; *)
                   if t2k = KType then 
                      (*  if subtype e ee t2 Ttop then *) (                     

                        let st = typecheck e ee s Ttop KType ff ih in ( 
                        let sk = typecheck e ee st KType KindOK ff ih in ( 
                          match (unrefine sk) with 
                          | KFun -> (let dom = nf_term (TDom st) e ee in
                                   let _ = typecheck e ee t dom KType ff ih in 
                                   let cod = nf_term (TCodom st) e ee in 
                                   validate_type e ee cod t2)
                          | KGen kg -> (  
                                     let _ =  typecheck e ee t kg KindOK ff ih in
                                     let TForall(x,dom,cod) = assertTForall (nf_term st e ee) in 
                                     let _  =  validate_type e ee kg dom in 
                                     let ee' = (x,nf_term t e ee)::ee in 
                                     let e' = (x,dom)::e in  ( 
                                     let tr = validate_type e' ee' (nf_term cod e' ee') t2 in (tr)
                                     ))
                                     )  
                     )) else if t2k = KindOK then 
                     ( 
                       let st = typecheck e ee s KType KindOK ff ih in  ( 
                       let _ = typecheck e ee st KindOK KindOK ff ih in ( (* print_term (App(s,t)); print_string " *** "; print_term st; print_string "  = st\n";*)
                       let KPi(x,dom,cod) = assertKPI ( nf_term st e ee ) in ( 
                       let _ =  typecheck e ee t dom KindOK ff ih in  
                       let ee' =  (x,t)::ee in (  
                       let e' =  (x,dom)::e in  ( 
                       let rtt = validate_type e' ee' (nf_term cod e' ee') t2 in ( rtt ) )
                     ))))) else 
                     failwith "Cannot apply this term as a function" )) in (arty)

and typecheckrefinement e ee t1 t2 ff =
  
        if ff then t2 else

        match t1 with 

 	|	Bool _ -> validate_type e ee TBool t2

        |       Id x -> validate_type e ee (find x e) t2

        |       Not a ->  let _ = typecheckrefinement e ee a TBool ff in validate_type e ee TBool t2

        |       And (a,b) ->  let _ = typecheckrefinement e ee a TBool ff in 
                              let _ = typecheckrefinement e ee b TBool ff in
                              (validate_type e ee TBool t2)

        |       Or (a,b) ->  let _ = typecheckrefinement e ee a TBool ff in 
                             let _ = typecheckrefinement e ee b TBool ff in
                             (validate_type e ee TBool t2)

        |       Implies (a,b) ->
                             let _ = typecheckrefinement e ee a TBool ff in 
                             let _ = typecheckrefinement e ee b TBool ff in
                             (validate_type e ee TBool t2)
        |       In (a, b) ->
                             let _ = typecheck e ee a KLab KindOK ff [] in 
                             let _ = typecheckrefinement e ee b LSet ff in (validate_type e ee TBool t2) 
        |       LabApart (a, b) ->
                             let _ = typecheckrefinement e ee a LSet ff in 
                             let _ = typecheckrefinement e ee b LSet ff in (validate_type e ee TBool t2) 
        |       LabUnion (a, b) -> 
                             let _ = typecheckrefinement e ee a LSet ff in 
                             let _ = typecheckrefinement e ee b LSet ff in (validate_type e ee LSet t2) 
        |       REmpty (a) ->
                             let _ = typecheck e ee a KRec KindOK ff [] in (validate_type e ee TBool t2) 
        |       LabSet (a) ->
                             let _ = typecheck e ee a KRec KindOK ff [] in (validate_type e ee LSet t2) 

        |       Equals (a, b, c) -> (
                            let (te,tk) =  ( try (* we need to infer types and kinds *)
                                          (typecheck e ee a Ttop KType ff [], KType) 
                                        with _ -> (typecheck e ee a KType KindOK ff [], KindOK)
                                      )
                            in 
                               let _ = typecheck e ee b te tk ff [] in
                               (validate_type e ee TBool t2))
;;

(*

         | REmpty(a)  -> (try
                         let _ = typecheck e ee a KType KindOK ff ih
                         in validate_type e ee TBool t2
                         with _ -> ( 
                            let at = typecheck e ee a Ttop KType ff ih
                            in let _ = typecheck e ee at KRec KindOK ff ih
                            in validate_type e ee TBool t2 ))

        |       Equals (a, b, c) -> print_string "EQ= "; print_term (Equals (a, b, c)); (* not OK *)
                             let tk = if subtype e ee c Ttop then KType else KindOK in
                             let at = typecheck e ee a c tk ff [] in  
                             let bt = typecheck e ee b c tk ff [] in
                             let ak' = (nf_term at e ee) in
                             let bk' = (nf_term bt e ee) in 
                             if ak' == bk' then (validate_type e ee TBool t2) else failwith "Kind mismatch in (t == u)"

        |       TLam (s,k,u) -> (
			let _ = typecheck e ee k KindOK (* use expected types *)
                        in try
                            ( let t = typecheck ((s,k)::e)  ((s,Id s)::ee) u Ttop (* eg., fun s::k -> 2 : forall X:k.int *)
                            in validate_type ee (TForall(s,(nf_term k ee),(nf_term t ((s,Id s)::ee)))) t2)
                           with _ ->
                            ( try                         
                              let t = typecheck ((s,k)::e)  ((s,Id s)::ee) u KType (* eg., fun s::k -> int :: Pi s:k.Type  *)
                              in ( 
                                   validate_type ee (KPi(s,(nf_term k ee), (nf_term t ((s,Id s)::ee))) ) t2)
                              with _ -> ( 
                                let _ = typecheck ((s,k)::e)  ((s,Id s)::ee) u KindOK (* eg.,Pi s::Type.Type :: OK  *)
                              in validate_type ee KindOK t2)
                            )
                           )

  |       App (s,t) -> (* rewrite based on expected type / kind *)
                     (try 
                        let st = typecheck e ee s Ttop 
                        and tt = typecheck e ee t Ttop in
                        let _ = typecheck e ee st KFun in
                        let dom = nf_term (TDom st) ee
                        and cod = nf_term (TCodom st) ee
                        in (
                             let _= validate_type ee tt dom
                             in validate_type ee cod t2)
                      with _ -> (
                         try
                           let st = typecheck e ee s Ttop in
                           let tt =  typecheck e ee t KType in
                           let _ = typecheck e ee st KType in
                           let TForall(x,dom,cod) = nf_term st ee in 
                           let _  =  validate_type ee tt dom in 
                           let ee' =  (x,nf_term t ee)::ee in 
                           validate_type ee' (nf_term cod ee') t2
                         with _ -> ( 
                             try 
                               let st = typecheck e ee s KType in 
                               let _ = typecheck e ee st KindOK in   
                               let KPi(x,dom,cod) = nf_term st ee in 
                               let tt =  typecheck e ee t dom in ( 
                               let _ =  validate_type ee dom tt in ( 
                               let ee' =  (x,nf_term t ee)::ee in 
                               validate_type ee' (nf_term cod ee') t2))
                            with _ -> (
                              failwith "No function in application lhs"
                             )
                        ) 
                     )
                   ) 

			let _ = typecheck e ee k KindOK (* use expected types *)
                        in try
                            ( let t = typecheck ((s,k)::e)  ((s,Id s)::ee) u Ttop (* eg., fun s::k -> 2 : forall X:k.int *)
                            in validate_type e ee (TForall(s,(nf_term k ee),(nf_term t ((s,Id s)::ee)))) t2)
                           with _ ->
                            (                          
                              let t = typecheck ((s,k)::e)  ((s,Id s)::ee) u KType (* eg., fun s::k -> int :: Pi s:k.Type  *)
                              in ( 
                                   validate_type e ee (KPi(s,(nf_term k ee), (nf_term t ((s,Id s)::ee))) ) t2)
                            )
                         )


let rec hnfrec t ee = 
  match t with
  | App(s,a) -> (match s with
                | Id f -> (match (find f ee) with 
                           | RecF (f,ty,b) -> (print_string ("XTREC\n\n");
                                match b with
                                | TLam (x,k,bf) -> ( 
                                                     let xn = new_id() in 
                                                     let bfr = ren_term bf [(x,Id xn)] in 
                                                 (f,ty,App(TLam (xn,k,bfr),a),a))
                                | _ -> ("",KindOK,KindOK,KindOK))                   
                           | _ -> ("",KindOK,KindOK,KindOK))

                | RecF (f,ty,b) -> (let nn = new_id () in (print_string ("XTREC="^nn^"\n"); failwith "NOT EXPECTED IN NF\n";
                                match b with
                                | TLam (x,k,bf) -> ( 
                                                     let xn = new_id() in 
                                                     let bfr = ren_term bf [(x,Id xn);(f, Id nn)] in 
                                                 (nn,ty,App(TLam (xn,k,bfr),a),a))

                                | _ -> ("",KindOK,KindOK,KindOK)))
                | App(s',a') -> (let (n,ty,b',ax) = hnfrec s ee in if b'=KindOK then (n, ty, App(s,a),ax) else (n, ty, App(b',a),ax))
                | _ ->  ("",KindOK,KindOK,KindOK)
                )
  | _ ->  ("",KindOK,KindOK,KindOK)
;;

let rec byIH s p ih =
  match ih with
  |   [] -> None
  |   (x,arg,t)::ih' -> if x = s then Some (x,arg,p,t) else byIH s p ih'
;;

let rec inductive t2 p ih = 
  match t2 with 
  | Id s -> byIH s p ih
  | App(f,u) -> inductive f u ih
  | _ -> None
;;

*)
